/*
 * Routines for interacting specifically with an Atmel 2051/4051
 */

#include <stdio.h>		/* used for string manipulation procedures */
#include <avr/io.h>		/* direct access to the AVRs IO ports/SFRs */
#include <avr/pgmspace.h>	/* for accessing strings in program memory */

#include <ctype.h>		/* to convert characters from upper to lowercase */

#include "avrutils.h"
#include "2051.h"

/*Iterate through the BurnIt chip memory and display on the 
 *    serial port, in ascii-coded hex
 *
 */

//static void print_ihex() {
//    printstr_p(PSTR("Displaying code in RAM\n"));
//    printstr_p(PSTR("ADDR :\tDATA\n"));
//    /*go thru all of memory*/
//    int16_t ix;
//    for(ix = 0; ix < CHIPMEMSIZE; ix++) {
//        if((ix % 32) == 0) {         /*line numbers and newlines */
//            printstr_p(PSTR("\n"));
//            char line_num[10];
//            sprintf(line_num, "%04X :\t", ix);
//            printstr(line_num);
//            
//        }
//
//        uint8_t curr_mem = chipMemory[ix];
//        printhex(curr_mem);
//
//    }
//
//}
/*
 *    Pulse XTAL1 on the ATMEL 2051/4051 to advance the programming
 *    memory location. 
 */
void atmel_pulse_xtal(void)
{
	CONTROL_PORT |= _BV(XTAL);
	_delay_us(1);
	CONTROL_PORT &= ~_BV(XTAL);
	_delay_us(1);
}

/*
 *    Write the data currently contained in memory to an Atmel 2051/4051
 *
 */
void atmel_write_chip(void)
{
	printstr_p(PSTR("\nWriting chip..."));
	CONTROL_PORT = 0x00;
	DATA_PORT = 0x00;
	_delay_us(10);
	CONTROL_PORT |= _BV(RST5) | _BV(P32);	//5v on RST, P3.2 high

	CONTROL_PORT &= ~(_BV(P33));	// L H H H    set mode bits.
	CONTROL_PORT |= _BV(P34) | _BV(P35) | _BV(P37);
	_delay_us(2);

	CONTROL_PORT |= _BV(RST12);	// Raise up to 12V (Enable programming)
	_delay_us(30);

	int16_t ix;
	for (ix = 0; ix < CHIPMEMSIZE; ix++) {
		if ((ix % 128) == 0) {	/*progress indicator */
			printstr_p(PSTR("."));
		}

		uint8_t curr_mem = chipMemory[ix];
		DATA_PORT = curr_mem;
		_delay_us(2);
		CONTROL_PORT &= ~(_BV(P32));
		_delay_us(10);
		CONTROL_PORT |= _BV(P32);

		loop_until_bit_is_clear(FEEDBACK_PIN, P31);
		loop_until_bit_is_set(FEEDBACK_PIN, P31);

		atmel_pulse_xtal();
		_delay_us(2);

	}
	_delay_ms(10);
	CONTROL_PORT &= ~(_BV(RST12));
	_delay_ms(10);
	CONTROL_PORT = 0x00;
	DATA_PORT = 0x00;
	printstr_p(PSTR(" OK. "));

}

/*
 * Erase an atmel 2051/4051. Sets all bytes to 0xFF.
 */
void atmel_clear_chip(void)
{
	printstr_p(PSTR("\nClearing chip... "));
	CONTROL_PORT = 0x00;
	CONTROL_PORT |= _BV(RST5) | _BV(P32);

	CONTROL_PORT |= _BV(P33);	//H L L L
	CONTROL_PORT &= ~(_BV(P34) | _BV(P35) | _BV(P37));
	_delay_us(10);
	CONTROL_PORT |= _BV(RST12);	// Raise up to 12V
	_delay_us(12);
	CONTROL_PORT &= ~(_BV(P32));
	_delay_ms(10);
	CONTROL_PORT |= _BV(P32);
	_delay_ms(10);
	CONTROL_PORT &= ~(_BV(RST12));
	_delay_ms(10);
	CONTROL_PORT = 0x00;
	printstr_p(PSTR("OK."));

}

/*
 *    Request a signature bytes from the AT2051/4051. 
 *    Returns 0 if a known signature is given
 *    1 if invalid signature is given
 */
int atmel_signature(void)
{
	printstr_p(PSTR("\nDetermining chip type..."));
	DATA_DDR = 0x00;
	CONTROL_PORT = 0x00;
	_delay_us(10);
	CONTROL_PORT |= _BV(RST5) | _BV(P32);

	CONTROL_PORT &= ~(_BV(P33) | _BV(P34) | _BV(P35) | _BV(P37));	// L L L L

	printstr_p(PSTR("0x"));
	_delay_us(2);		//time until data is valid

	uint8_t manufact_byte;
	manufact_byte = DATA_PIN;
	printhex(manufact_byte);

	printstr_p(PSTR(" 0x"));
	atmel_pulse_xtal();
	_delay_us(2);

	uint8_t part_byte;
	part_byte = DATA_PIN;
	printhex(part_byte);

	CONTROL_PORT = 0x00;
	DATA_DDR = 0xFF;
	DATA_PORT = 0x00;

	if (manufact_byte == 0x1e && part_byte == 0x21) {
		printstr_p(PSTR(". Atmel 2051. "));
		return 0;
	} else if (manufact_byte == 0x1e && part_byte == 0x41) {
		printstr_p(PSTR(". Atmel 4051. "));
		return 0;
	} else {
		printstr_p(PSTR("\nUnknown Part...Check Connections"));
		return 1;
	}

}

/* 
 *    View the Data currently on the chip
 */
void atmel_view(void)
{

	DATA_DDR = 0x00;
	CONTROL_PORT = 0x00;
	_delay_us(10);
	CONTROL_PORT |= _BV(RST5) | _BV(P32);

	CONTROL_PORT &= ~(_BV(P33) | _BV(P34));	// L L H H
	CONTROL_PORT |= _BV(P35) | _BV(P37);
	_delay_us(2);

	int16_t ix;
	for (ix = 0; ix < CHIPMEMSIZE; ix++) {
		if ((ix % 32) == 0) {	/*line numbers and newlines */
			printstr_p(PSTR("\n"));
			char line_num[10];
			sprintf(line_num, "%04X :\t", ix);
			printstr(line_num);

		}

		uint8_t curr_mem = DATA_PIN;
		printhex(curr_mem);
		atmel_pulse_xtal();
		_delay_us(2);

	}
	CONTROL_PORT = 0x00;
	DATA_DDR = 0xFF;
	DATA_PORT = 0x00;

}

/*
 *    Checks to see if the data in the internal representation of
 *    the chip's memory matches the data on the currently inserted chip
 */
void atmel_verify(void)
{
	printstr_p(PSTR("\nVerifying program data..."));
	DATA_DDR = 0x00;
	CONTROL_PORT = 0x00;
	_delay_us(10);
	CONTROL_PORT |= _BV(RST5) | _BV(P32);

	CONTROL_PORT &= ~(_BV(P33) | _BV(P34));	// L L H H
	CONTROL_PORT |= _BV(P35) | _BV(P37);
	_delay_us(2);

	int16_t ix;
	for (ix = 0; ix < CHIPMEMSIZE; ix++) {
		if ((ix % 128) == 0) {	/*progress indicator */
			printstr_p(PSTR("."));
		}
		uint8_t curr_mem = DATA_PIN;
		if (curr_mem != chipMemory[ix]) {
			printstr_p(PSTR("Verification Failed!"));
			CONTROL_PORT = 0x00;
			DATA_DDR = 0xFF;
			DATA_PORT = 0x00;
			return;
		}
		atmel_pulse_xtal();
		_delay_us(2);

	}
	CONTROL_PORT = 0x00;
	DATA_DDR = 0xFF;
	DATA_PORT = 0x00;
	printstr_p(PSTR(" OK."));

}

/*
 *    Go through the steps of programming:
 *        check the signature, if invalid then do nothing
 *        if valid, clear the chip, write the chip and then verify it
 */
void atmel_program(void)
{
	printstr_p(PSTR("\nPreparing to program Atmel 2051/4051..."));
	if (atmel_signature() == 0) {
		atmel_clear_chip();
		atmel_write_chip();
		atmel_verify();
	}
	//verify chip
}
