//JEDEC and GAL algorithms
//Taken from GALBlast and ported to AVR without interface

#include <stdio.h>		/* used for string manipulation procedures */
#include <avr/io.h>		/* direct access to the AVRs IO ports/SFRs */
#include <avr/pgmspace.h>	/* for accessing strings in program memory */
#include <ctype.h>		/* to convert from upper to lowercase */
#include <string.h>

#include "avrutils.h"
#include "gal.h"

//Programming Pulse time.  Learned from PES
static int pulse = 0;

//Erase Pulse time. Learned from PES
static int erase = 0;

//Programming Voltage from PES.
static int vpp = 0;

//Array mapping PES duration values to milliseconds
static int duration[16] =
    { 1, 2, 5, 10, 20, 30, 40, 50, 60, 70, 80, 90, 100, 200, 0, 0 };

static unsigned char fusemap[737];

unsigned char pes[12];

static int cfg22V10[] = {
	5809, 5808,
	5811, 5810,
	5813, 5812,
	5815, 5814,
	5817, 5816,
	5819, 5818,
	5821, 5820,
	5823, 5822,
	5825, 5824,
	5827, 5826
};

GALINFO galinfo[] = {
	{UNKNOWN, 0x00, 0x00, "unknown", 0, 0, 0, 0, 0,
	 0, 0, 0, 0, 0, 8, 0, 0, 0},
	{GAL22V10, 0x48, 0x49, "GAL22V10", 5892, 24, 44,
	 132, 44, 5828, 8, 61, 60, 58, 10, 16, cfg22V10,
	 sizeof(cfg22V10) / sizeof(int)},
};

/*
 * SetAddr(uint8_t addr)
 * Sets the address lines of the GAL22V10 to the specified value
 */
static void SetAddr(uint8_t addr)
{
	//Unfortunately, these address lines are scattered across different ports, so each bit has to be
	//done individually.
	if (addr & _BV(0)) {
		GALPA0PORT |= _BV(GALPA0);
	} else {
		GALPA0PORT &= ~_BV(GALPA0);
	}
	if (addr & _BV(1)) {
		GALPA1PORT |= _BV(GALPA1);
	} else {
		GALPA1PORT &= ~_BV(GALPA1);
	}
	if (addr & _BV(2)) {
		GALPA2PORT |= _BV(GALPA2);
	} else {
		GALPA2PORT &= ~_BV(GALPA2);
	}
	if (addr & _BV(3)) {
		GALPA3PORT |= _BV(GALPA3);
	} else {
		GALPA3PORT &= ~_BV(GALPA5);
	}
	if (addr & _BV(4)) {
		GALPA4PORT |= _BV(GALPA4);
	} else {
		GALPA4PORT &= ~_BV(GALPA4);
	}
	if (addr & _BV(5)) {
		GALPA5PORT |= _BV(GALPA5);
	} else {
		GALPA5PORT &= ~_BV(GALPA5);
	}
}

/*
 * SetEDIT()
 * sets the value of the EDIT line on the GAL
 */
static void SetEDIT(int setEdit)
{
	if (setEdit) {
		EDITPORT |= _BV(EDIT);
	} else {
		EDITPORT &= ~(_BV(EDIT));
	}
}

/*
 * Sets the drivers to be either on or off
 */
static void SetDrivers(int setDrivers)
{
	if (setDrivers) {
		//Drive everything low except for EDIT itself and STB on PortB
		PORTA = 0x00;;
		PORTB = _BV(STB);
		PORTC = 0x00;
		EDITPORT &= _BV(EDIT);	//Leave EDIT alone, but put everything else low
		DDRA = PORTA_GAL_DDR;
		DDRB = PORTB_GAL_DDR;
		DDRC = PORTC_GAL_DDR;
		DDRD = PORTD_GAL_DDR;
	} else {
		//Convert all back to inputs except the high voltage bits.  They keep driving
		DDRA = 0x00;
		DDRB = _BV(PB1);	//This is to keep the 2051 high voltage driver from being silly
		DDRC = 0x00;
		EDITDDR = _BV(EDIT);

		//Now enable pullups except on ultra high voltages
		PORTA = 0xFF;
		PORTB = ~(_BV(PB1));
		PORTC = 0xFF;
		//Enable all pullups except on EDIT.  Just leave it alone
		PORTD |= ~(_BV(EDIT));
	}
}

static void SetPV(int setPV)
{
	if (setPV) {
		GALCTRLPORT |= _BV(PV);
	} else {
		GALCTRLPORT &= ~(_BV(PV));
	}
}

static void SetSCLK(int setSCLK)
{
	if (setSCLK) {
		GALCTRLPORT |= _BV(SCLK);
	} else {
		GALCTRLPORT &= ~(_BV(SCLK));
	}
}

//Enters programming mode.  This will take care of initializing output pins
//Mode determines whether we are doing a READ or some kind of write.  The programming voltage is different
static char TurnOn(int mode)
{
	char writeorerase;

	//First decide if this turn on will involve writing.
	if (mode == WRITEGAL || mode == ERASEGAL || mode == ERASEALL ||
	    mode == BURNSECURITY || mode == WRITEPES || mode == VPPTEST) {
		writeorerase = 1;
	} else {
		writeorerase = 0;
	}
	//Pins should not be driving, but make sure
	SetDrivers(0);

	_delay_us(20);

	//Turn on programming/reading voltage only
	SetEDIT(1);

	_delay_ms(100);

	_delay_ms(10);

	//Now all pins in programming mode, so drivers can come on
	//This puts all VIL LOW and all GAL INPUTS to LOW except STB
	SetDrivers(1);

	_delay_ms(20);

	if (writeorerase) {
		SetPV(1);
		_delay_ms(10);
	} else {
		SetPV(0);
		_delay_ms(10);
	}
	return 1;
}

//Turns off the programming mode
static void TurnOff(void)
{
	_delay_us(200);

	//Disable the drivers.
	SetDrivers(0);

	_delay_ms(10);

	//Turn off edit now
	SetEDIT(0);

	_delay_us(20);
}

void SendBit(int bit)
{
	if (bit) {
		GALCTRLPORT |= _BV(SDIN);
	} else {
		GALCTRLPORT &= ~(_BV(SDIN));
	}

	_delay_us(2);
	SetSCLK(1);
	_delay_us(2);
	SetSCLK(0);
	_delay_us(2);
}

void SendBits(int n, int bit)
{
	while (n-- > 0)
		SendBit(bit);
}

void Strobe(int msec)
{
	//_delay_ms(3);
	_delay_ms(1);
	GALCTRLPORT &= ~(_BV(STB));
	_delay_ms(msec);
	GALCTRLPORT |= _BV(STB);
	//_delay_ms(3);
	_delay_ms(1);
}

void SendAddress(int n, int row)
{
	while (n-- > 0) {
		SendBit(row & 1);
		row >>= 1;
	}
}

//Selects the row we want to talk to
void StrobeRow(int row)
{
	SetAddr(0);
	SendBits(132, 0);	//Sends 132 zeroes (because 132 bits per row in GAL22V10)
	SendAddress(6, row);	//Now send address of the row we want
	//The 6 is for how many bits are in the row

	Strobe(1);		//Strobe the pin 
}

unsigned char ReceiveBit(void)
{
	unsigned char bit;

	bit = (SDOUTPIN & _BV(SDOUT) ? 1 : 0);
	_delay_us(2);
	SetSCLK(1);
	_delay_us(2);
	SetSCLK(0);
	_delay_us(2);
	return bit;
}

//Parses the PES information retreived from the GAL
//This contains information on programming voltage, programming time, and other things
void ParsePES(int displayPES)
{
	int algo = pes[1] & 0x0F;

	if (algo == 5) {	//Algorithm is specified
		erase = (25 << ((pes[4] >> 2) & 7)) / 2;
		pulse =
		    duration[((((unsigned)pes[5] << 8) | pes[4]) >> 5) & 15];
		vpp = 2 * ((pes[5] >> 1) & 31) + 20;
	} else {
		erase = (pes[3] == NATIONAL ? 50 : 100);
		switch (algo) {
		case 0:
			vpp = 66;	// 16.5V
			pulse = 10;
			break;
		case 1:
			vpp = 63;	// 15.75V
			pulse = 100;
			break;
		case 2:
			vpp = pes[3] == NATIONAL ? 60 : 58;	// 15/14.5V
			pulse = 40;
			break;
		case 3:
			vpp = 56;	// 14V
			pulse = 100;
			break;
		}
	}
	//I am hardwiring this to be a long time so that the chip will work even if the PES is toasted.
	pulse = 100;
	erase = 200;

	//Note: I do not want to compile in software floating point arithmetic
	//for this. Since I am just dividing by two, I can do it with bit shifts
	if (displayPES) {
		char s[100];
		int realvpp = vpp / 4;
		char isQuarter = vpp & 0x01;	//See if dividing by two truncates
		char isHalf = vpp & 0x02 ? 1 : 0;
		char fraction = isQuarter * 25 + isHalf * 50;
		sprintf(s, "\nPES says Vpp = %d.%d,Pulse "
			"time = %d ms", realvpp, fraction, pulse);
		printstr(s);
		sprintf(s, "\nPES: %d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d", pes[0],
			pes[1], pes[2], pes[3], pes[4], pes[5], pes[6], pes[7],
			pes[8], pes[9], pes[10], pes[11]);
		printstr(s);
	}
	return;
}

//Reads the Programmer's Electronic Signature to make sure
//this is a GAL22V10 and we know the programming
//algorithm
void ReadPES()
{
	int bitmask, byte;

	TurnOn(READPES);

	StrobeRow(58);		//Select row 58, the PES on GAL22V10
	for (byte = 0; byte < 10; byte++) {	//The PES is 10 bytes long.  Get them
		pes[byte] = 0;
		for (bitmask = 0x1; bitmask <= 0x80; bitmask <<= 1) {	//Get each bit and
			//assemble into PES
			if (ReceiveBit())
				pes[byte] |= bitmask;
		}
	}

	TurnOff();
}

int ReadGAL(char verify)
{
	int row, bit;
	short fuseindex;
	int fusebit;

	char s[7];
	int errorfound = 0;

	char tempfusemap = 0;

	TurnOn(READGAL);

	for (row = 0; row < galinfo[GAL22V10].rows; row++) {
		StrobeRow(row);
		for (bit = 0; bit < galinfo[GAL22V10].bits; bit++) {
			fuseindex = (galinfo[GAL22V10].rows * bit + row) / 8;
			fusebit = (galinfo[GAL22V10].rows * bit + row) % 8;
			if (verify) {
				tempfusemap = ReceiveBit()?
				    tempfusemap | _BV(fusebit) :
				    tempfusemap & ~(_BV(fusebit));
				if ((tempfusemap & _BV(fusebit)) !=
				    (fusemap[fuseindex] & _BV(fusebit))) {
					sprintf(s, "\n%hd",
						fuseindex * 8 + fusebit);
					printstr(s);
					errorfound = 1;
				}
			} else {
				fusemap[fuseindex] = ReceiveBit()?
				    fusemap[fuseindex] | _BV(fusebit) :
				    fusemap[fuseindex] & ~(_BV(fusebit));
			}
		}
	}

	// UES
	StrobeRow(galinfo[GAL22V10].uesrow);
	for (bit = 0; bit < galinfo[GAL22V10].uesbytes * 8; bit++) {
		fuseindex = (galinfo[GAL22V10].uesfuse + bit) / 8;
		fusebit = (galinfo[GAL22V10].uesfuse + bit) % 8;
		if (verify) {
			tempfusemap = ReceiveBit()?
			    tempfusemap | _BV(fusebit) :
			    tempfusemap & ~(_BV(fusebit));
			if ((tempfusemap & _BV(fusebit)) !=
			    (fusemap[fuseindex] & _BV(fusebit))) {
				sprintf(s, "\n%hd", fuseindex * 8 + fusebit);
				printstr(s);
				errorfound = 1;
			}
		} else {
			fusemap[fuseindex] = ReceiveBit()?
			    fusemap[fuseindex] | _BV(fusebit) :
			    fusemap[fuseindex] & ~(_BV(fusebit));
		}
	}

	// CFG
	SetAddr(galinfo[GAL22V10].cfgrow);
	Strobe(2);
	for (bit = 0; bit < galinfo[GAL22V10].cfgbits; bit++) {
		fuseindex = (galinfo[GAL22V10].cfg[bit]) / 8;
		fusebit = (galinfo[GAL22V10].cfg[bit]) % 8;
		if (verify) {
			tempfusemap = ReceiveBit()?
			    tempfusemap | _BV(fusebit) :
			    tempfusemap & ~(_BV(fusebit));
			if ((tempfusemap & _BV(fusebit)) !=
			    (fusemap[fuseindex] & _BV(fusebit))) {
				sprintf(s, "\n%hd", fuseindex * 8 + fusebit);
				printstr(s);
				errorfound = 1;
			}
		} else {
			fusemap[fuseindex] = ReceiveBit()?
			    fusemap[fuseindex] | _BV(fusebit) :
			    fusemap[fuseindex] & ~(_BV(fusebit));
		}
	}

	TurnOff();
	return errorfound;
}

void PrintFuses(void)
{
	int i;
	char s[8];
	for (i = 0; i < sizeof(fusemap); i++) {
		if (i % 32 == 0) {
			printstr_p(PSTR("\n"));
		}
		sprintf(s, "%X ", fusemap[i]);
		printstr(s);
	}
}

void EraseGAL(void)
{
	if (TurnOn(ERASEGAL)) {
		SetAddr(61);
		Strobe(erase);
		TurnOff();
	}
}

//Write the current fusemap to the GAL in the socket using data from PES
void WriteGAL(int ones)
{
	int row, bit;
	short fuseindex;
	int fusebit;

	if (TurnOn(WRITEGAL)) {
		SetAddr(0);
		for (row = 0; row < galinfo[GAL22V10].rows; row++) {
			for (bit = 0; bit < galinfo[GAL22V10].bits; bit++) {
				fuseindex =
				    (galinfo[GAL22V10].rows * bit + row) / 8;
				fusebit =
				    (galinfo[GAL22V10].rows * bit + row) % 8;
				if (ones) {
					SendBit(1);
				} else {
					SendBit(fusemap[fuseindex] &
						_BV(fusebit) ? 1 : 0);
				}
			}
			SendAddress(6, row);
			GALCTRLPORT &= ~(_BV(SDIN));
			Strobe(pulse);
		}
		// UES
		for (bit = 0; bit < galinfo[GAL22V10].uesbytes * 8; bit++) {
			fuseindex = (galinfo[GAL22V10].uesfuse + bit) / 8;
			fusebit = (galinfo[GAL22V10].uesfuse + bit) % 8;
			if (ones) {
				SendBit(1);
			} else {
				SendBit(fusemap[fuseindex] & _BV(fusebit) ? 1 :
					0);
			}
		}
		if (galinfo[GAL22V10].uesbytes * 8 < galinfo[GAL22V10].bits) {
			SendBits(galinfo[GAL22V10].bits -
				 galinfo[GAL22V10].uesbytes * 8, 0);
		}
		SendAddress(6, galinfo[GAL22V10].uesrow);
		GALCTRLPORT &= ~(_BV(SDIN));
		Strobe(pulse);
		// CFG
		SetAddr(galinfo[GAL22V10].cfgrow);
		for (bit = 0; bit < galinfo[GAL22V10].cfgbits; bit++) {
			fuseindex = (galinfo[GAL22V10].cfg[bit]) / 8;
			fusebit = (galinfo[GAL22V10].cfg[bit]) % 8;
			if (ones) {
				SendBit(1);
			} else {
				SendBit(fusemap[fuseindex] & _BV(fusebit) ? 1 :
					0);
			}
		}
		GALCTRLPORT &= ~(_BV(SDIN));
		Strobe(pulse);
	}
	TurnOff();
}

//Computes the checksum of the fusemap
//argument is the size of the fusemap.  For GAL22V10, should be
//5892, but this is checked before CheckSum is called
static unsigned short CheckSum(int n)
{
	unsigned short c, e;
	long a;
	int i;

	unsigned short fusemapindex = 0;
	unsigned char bitindex = 0;

	c = e = 0;
	a = 0;
	for (i = 0; i < n; i++) {
		e++;
		if (e == 9) {
			e = 1;
			a += c;
			c = 0;
		}
		c >>= 1;
		fusemapindex = i / 8;
		bitindex = i % 8;
		if (fusemap[fusemapindex] & _BV(bitindex))
			c += 0x80;
	}
	return (unsigned short)((c >> (8 - e)) + a);
}

//Parses a JEDEC file being transmitted over the serial port
//Uses a messy state machine to do it.  Returns 1 if successful and 0 if an error
int readJEDEC()
{
	//int i,n,type,checksumpos,address,pins,lastfuse,state;

	// 0=outside JEDEC, 1=skipping comment or unknown, 2=read command
	// Other states were undocumented.
	//Looking at the JEDEC standard, a transmission starts with hex 0x02 
	//and ends with hex 0x03
	//Every useful line starts with a '*' character, so the 
	//state machine goes to 2

	short address = 0;

	//Used because the fusemap must be packed for space reasons.
	short fuseaddress = 0;
	char bitlocation = 0;

	//This is the 16-bit sum of the entire transmission
	unsigned short xmitchecksum = 0;

	char state = 0;
	char security = 0;
	short checksum = 0;
	char pins = 0;
	short lastfuse = 0;

	unsigned char receivedChar;

	printstr_p(PSTR("\n>"));

	//Wait until we get START byte
	while (getchr() != 0x02) {
		;
	}
	xmitchecksum = 0x02;
	//We got it.  Now process the file
	while ((receivedChar = getchr())) {
		xmitchecksum = (xmitchecksum + receivedChar) & 0xFFFF;
		//First check for end
		if (receivedChar == 0x03) {
			//We're done
			break;
		} else if (receivedChar == '*') {
			printstr_p(PSTR("."));
			state = 2;
		} else
			switch (state) {
			case 2:	//This char defines what the line is
				if (!isspace(receivedChar)) {
					switch (receivedChar) {
					case 'L':	//Fuse List
						address = 0;
						state = 3;
						break;
					case 'F':	//Default Fuse State
						state = 5;
						break;
					case 'G':	//Security Fuse
						state = 13;
						break;
					case 'Q':	//Specifies features of GAL
						state = 7;
						break;
					case 'C':	//Fuse Checksum
						state = 14;
						break;
					default:
						state = 1;
					}
				}
				break;
			case 3:	//Getting first digit of Fuse List
				if (!isdigit(receivedChar))
					return 0;
				address = receivedChar - '0';
				state = 4;
				break;
			case 4:	//Getting remaining digits of Fuse List until Space received
				if (isspace(receivedChar)) {
					state = 6;
				} else if (isdigit(receivedChar)) {
					address =
					    10 * address + (receivedChar - '0');
				} else {
					return 0;
				}
				break;
			case 5:	//Default Fuse state command "format" the fusemap
				if (isspace(receivedChar))
					break;	// ignored
				if (receivedChar == '0' || receivedChar == '1') {
					memset(fusemap, receivedChar - '0',
					       sizeof(fusemap));
				} else {
					return 0;
				}
				state = 1;
				break;
			case 6:	//Reading in fuses from Fuse List entry until next '*'
				if (isspace(receivedChar))
					break;	// ignore spaces
				if (receivedChar == '0' || receivedChar == '1') {

					//Divide by 8 to get index into packed fusemap array
					fuseaddress = address / 8;
					//Use modulo to get which bit in that cell to read
					bitlocation = address % 8;

					//Finally, write the bit into that cell
					if (receivedChar == '0') {
						fusemap[fuseaddress] &=
						    ~(_BV(bitlocation));
					} else {
						fusemap[fuseaddress] |=
						    _BV(bitlocation);
					}
					address++;
				} else {
					return 0;
				}
				break;
			case 7:	//Get configuration information.  Currently ignored
				if (isspace(receivedChar))
					break;	// ignored
				if (receivedChar == 'P') {
					pins = 0;
					state = 8;
				} else if (receivedChar == 'F') {
					lastfuse = 0;
					state = 9;
				} else
					state = 2;
				break;
			case 8:	//Setting expected number of pins
				if (isspace(receivedChar))
					break;	// ignored
				if (!isdigit(receivedChar))
					return 0;
				pins = receivedChar - '0';
				state = 10;
				break;
			case 9:	//Setting expected number of fuses
				if (isspace(receivedChar))
					break;	// ignored
				if (!isdigit(receivedChar))
					return 0;
				lastfuse = receivedChar - '0';
				state = 11;
				break;
			case 10:	//Getting remaining digits of number of pins
				if (isdigit(receivedChar)) {
					pins = 10 * pins + (receivedChar - '0');
				} else if (isspace(receivedChar)) {
					state = 12;
				} else
					return 0;
				break;
			case 11:	//Getting remaining digits of number of fuses
				if (isdigit(receivedChar)) {
					lastfuse =
					    10 * lastfuse + (receivedChar -
							     '0');
				} else if (isspace(receivedChar)) {
					state = 12;
				} else
					return 0;
				break;
			case 12:	//Ensuring that there is whitespace after setting config?
				if (!isspace(receivedChar))
					return 0;
				break;
			case 13:	//Security Fuse setting
				if (isspace(receivedChar))
					break;	// ignored
				if (receivedChar == '0' || receivedChar == '1') {
					security = receivedChar - '0';
				} else {
					return 0;
				}
				state = 1;
				break;
			case 14:	//Get the checksum first digit
				if (isspace(receivedChar))
					break;	// ignored
				if (isdigit(receivedChar)) {
					checksum = receivedChar - '0';
				} else if (toupper(receivedChar) >= 'A' &&
					   toupper(receivedChar) <= 'F') {
					checksum =
					    toupper(receivedChar) - 'A' + 10;
				} else
					return 0;
				state = 15;
				break;
			case 15:	//Get the remaining digits until a space
				if (isdigit(receivedChar)) {
					checksum =
					    16 * checksum + receivedChar - '0';
				} else if (toupper(receivedChar) >= 'A'
					   && toupper(receivedChar) <= 'F') {
					checksum =
					    16 * checksum +
					    toupper(receivedChar) - 'A' + 10;
				} else if (isspace(receivedChar)) {
					state = 2;
				} else
					return 0;
				break;
			}
	}			//Ends for loop

	//Now get the transmit checksum
	unsigned short tempxmitcheck = 0;
	tempxmitcheck = ascii_to_bin(getchr());
	tempxmitcheck = (tempxmitcheck << 4) | ascii_to_bin(getchr());
	tempxmitcheck = (tempxmitcheck << 4) | ascii_to_bin(getchr());
	tempxmitcheck = (tempxmitcheck << 4) | ascii_to_bin(getchr());

	/*
	 * Note we are not checking the transmission checksum. 
	 * It is broken if the JED file
	 * moves between Unix and Windows or if it is mangled in Hyperterminal because it includes 
	 * line feeds and carriage returns as meaningful elements.
	 * If the file is moved as a text file, line endings get 
	 * mangled and invalidate the checksum.
	 * Besides, there is already a fuse checksum, so this feels superfluous.
	 */

	//Make sure that a GAL22V10 was used
	if (lastfuse != 5892) {
		printstr_p(PSTR("\nThis JEDEC has the wrong number of fuses or"
				" no fuse declaration.  Is it for a GAL22V10?"));
		return 0;
	}
	if (pins != 24) {
		printstr_p(PSTR
			   ("\nThis JEDEC has the wrong number of pins or no"
			    " pin declaration.  Is it for a GAL22V10?"));
		return 0;
	}
	//Now compute fuse checksum using the known fuse size
	//I will demand that a checksum be present
	if (checksum == 0) {
		printstr_p(PSTR("\nNo Fuse Checksum present in JEDEC file.  "
				"Check compiler settings"));
		return 0;
	}
	if (checksum != CheckSum(lastfuse)) {
		printstr_p(PSTR("\nChecksum Wrong"));
		return 0;
	}
	if (security == 1) {
		printstr_p(PSTR
			   ("\nYou have chosen to program the Security Fuse in this JEDEC."));
		printstr_p(PSTR
			   ("\nBecause this will prevent reading out the fuse map, this GAL"));
		printstr_p(PSTR
			   ("\nwill give a verification error after programming."));
		printstr_p(PSTR
			   ("\nSimply clear or reprogram the GAL to clear the security fuse"));
	}
	return 1;
}
